/*
 * *** Please do not edit ! ***
 * @(#) Leash_1.cs
 * @procedure leash/1 in interpreter.pl
 */

/*
 * @version P# 0.1, on March 18 2002;  Prolog Cafe 0.44, on November 12 1999
 * @author Mutsunori Banbara (banbara@pascal.seg.kobe-u.ac.jp)
 * @author Naoyuki Tamura    (tamura@kobe-u.ac.jp)
 * Modified by Jonathan Cook (jjc@dcs.ed.ac.uk)
 */
namespace JJC.Psharp.Predicates {

using JJC.Psharp.Lang;
using JJC.Psharp.Lang.Resource;
using Predicates = JJC.Psharp.Predicates;
using Resources = JJC.Psharp.Resources;

public class Leash_1 : Predicate {
    static internal readonly Predicate Leash_1_1 = new Predicates.Leash_1_1();
    static internal readonly Predicate Leash_1_2 = new Predicates.Leash_1_2();
    static internal readonly Predicate Leash_1_sub_1 = new Predicates.Leash_1_sub_1();

    public Term arg1;

    public Leash_1(Term a1, Predicate cont) {
        arg1 = a1; 
        this.cont = cont;
    }

    public Leash_1(){}
    public override void setArgument(Term[] args, Predicate cont) {
        arg1 = args[0]; 
        this.cont = cont;
    }

    public override Predicate exec( Prolog engine ) {
        engine.aregs[1] = arg1;
        engine.cont = cont;
        return call( engine );
    }

    public virtual Predicate call( Prolog engine ) {
        engine.setB0();
        return engine.jtry(Leash_1_1, Leash_1_sub_1);
    }

    public override int arity() { return 1; }

    public override string ToString() {
        return "leash(" + arg1 + ")";
    }
}

sealed class Leash_1_sub_1 : Leash_1 {

    public override Predicate exec( Prolog engine ) {
        return engine.trust(Leash_1_2);
    }
}

sealed class Leash_1_1 : Leash_1 {

    public override Predicate exec( Prolog engine ) {
        Term a1, a2;
        Predicate p1, p2, p3, p4;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        a2 = engine.makeVariable();
        p1 = new Predicates.dollar_leash_1(a1, cont);
        p2 = new Predicates.dollar_cut_1(a2, p1);
        p3 = new Predicates.dollar_checkLeash_1(a1, p2);
        p4 = new Predicates.dollar_nonvar_1(a1, p3);
        return new Predicates.dollar_getLevel_1(a2, p4);
    }
}

sealed class Leash_1_2 : Leash_1 {
    static internal readonly SymbolTerm f1 = SymbolTerm.makeSymbol("$leash_mode", 1);
    static internal readonly SymbolTerm f3 = SymbolTerm.makeSymbol("domain", 2);
    static internal readonly SymbolTerm s4 = SymbolTerm.makeSymbol("atom");
    static internal readonly SymbolTerm f6 = SymbolTerm.makeSymbol("leash", 1);
    static internal readonly IntegerTerm s8 = new IntegerTerm(1);

    public override Predicate exec( Prolog engine ) {
        Term a1, a2, a3, a4, a5, a6;
        Predicate p1;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        a2 = engine.makeVariable();
        Term[] h2 = {a2};
        a3 = new StructureTerm(f1, h2);
        a4 = engine.makeVariable();
        Term[] h5 = {s4, a4};
        a5 = new StructureTerm(f3, h5);
        Term[] h7 = {a1};
        a6 = new StructureTerm(f6, h7);
        p1 = new Predicates.dollar_illarg_3(a5, a6, s8, cont);
        return new Predicates.Findall_3(a2, a3, a4, p1);
    }
}
}

