/*
 * @(#)dollar_call_1.cs
 * $B=R8l(B $call/1
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;

	public sealed class dollar_call_1 : Predicate {
		public static SymbolTerm SYM_true = SymbolTerm.makeSymbol("true");
		public static SymbolTerm SYM_fail = SymbolTerm.makeSymbol("fail");

		Term arg1;

		public dollar_call_1(Term a1, Predicate cont) {
			arg1 = a1;
			this.cont = cont;
		}
		public dollar_call_1() {}
		public override void setArgument(Term[] args, Predicate cont) {
			arg1 = args[0];
			this.cont = cont;
		}

		public override Predicate exec( Prolog engine ) {
			string pred_name = null;
			SymbolTerm functor = null;
			Term[] args = null;
			int arity;

			Term t = arg1.dereference();

			if ( t.isSymbol() ){
				functor = (SymbolTerm) t;
				arity = 0;
				if ( functor.Equals(SYM_true) ) 
					return cont;
				else if ( functor.Equals(SYM_fail) ) 
					return engine.fail();
				else
					pred_name = functor.name;
			}
			else if ( t.isStructure() ){
				functor = ((StructureTerm) t).functor;
				args  = ((StructureTerm) t).args;
				arity = ((StructureTerm) t).arity;
				pred_name = functor.name;      
			}
			else if ( t.isVariable() ){
				return engine.abort("{INSTANTIATION ERROR: " + this +
					" -arg 1}\n");
				//	    return null;
			}
			else {
				return engine.abort("{TYPE ERROR: " + this +
					" -arg 1: expected goal, found " +
					t + "}\n");
				//	    return null;
			}

			Predicate pred;
			// System.Type pred_class = null;
			string class_name;

			class_name = getClassName(pred_name, arity);

			System.Type clazz;
			try {
				clazz = AssemblyList.GetAnyType( class_name );	
				pred = (Predicate)System.Activator.CreateInstance( clazz );
				pred.setArgument( args, cont );
			} catch( System.Exception ) {
				// System.Console.WriteLine( "Failed to find: pred: " + pred_name + "; class: " + class_name );
				return engine.fail();
			}
			return pred;

		}

		public override int arity() { return 1; }

		public override string ToString() {
			return "$call(" + arg1 + ")";
		}

		/* This has been changed to generate the new style names. 
		 * JJC */

		public string getClassName( string s, int arity ) {
			char[] ch;
			int len;
			string s1 = "";
			
			ch = s.ToCharArray( );
			len = ch.Length;

			if( System.Char.IsLower( ch[ 0 ] ) )
				s1 += System.Char.ToUpper( ch[ 0 ] );
			else if( ch[ 0 ] == '_' )
				s1 += "__";
			else if( System.Char.IsDigit( ch[ 0 ] ) || 
					 System.Char.IsUpper( ch[ 0 ] ) )
				s1 += "_" + ch[ 0 ];
			else 
				s1 += Token.getReplacement( (int)ch[ 0 ] );

			int i = 1;
			while( i < len ) {
				if( i == len - 1 && ch[ i ] == '_' ) {
					s1 += "_underscore";
					break;
				} else if( System.Char.IsLower( ch[ i ] ) ||
					System.Char.IsDigit( ch[ i ] ) )
					s1 += ch[ i ];
				else if( System.Char.IsUpper( ch[ i ] ) )
					s1 += "_" + ch[ i ];
				else if( ch[ i ] == '_' ) {
					i++;
					if( System.Char.IsUpper( ch[ i ] ) ||
						System.Char.IsDigit( ch[ i ] ) )
						s1 += "__" + ch[ i ];
					else if( System.Char.IsLower( ch[ i ] ) )
						s1 += System.Char.ToUpper( ch[ i ] );
					else if( ch[ i ] == '_' )
						s1 += "__underscore";
					else 
						s1 += "__" + Token.getReplacement( (int)ch[ i ] );
				} else
					s1 += "_" + Token.getReplacement( (int)ch[ i ] );
				i++;
			}

			return "JJC.Psharp.Predicates." + s1 + "_" + arity;
		}


		/*
		public string getClassName(string s, int arity) {
			char[] ch;
			int len;
			string s1 = "";

			ch = s.ToCharArray();
			len = ch.Length;
			for (int i=0; i<len; i++) {
				if ( ch[i] == '$') {
					s1 = s1 + "36"; // ch[i];
				}
				else if ( Token.isSymbol((int)ch[i]) || 
					Token.isSolo((int)ch[i])   || ch[i] == ',') {
					s1 = s1 + (int)ch[i];
				} else {
					s1 = s1 + ch[i];
				}
			}
			if (System.Char.IsDigit(s1.ToCharArray()[0]))
				return "PRED_I" + s1 + "_" + arity;
			else
				return "PRED_" + s1 + "_" + arity;
		}
	*/
	}
	
}