/*
 * dollar_functor_3.cs
 * $B=R8l(B $functor/3
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;

	public sealed class dollar_functor_3 : Predicate {
		public static SymbolTerm SYM_dot = SymbolTerm.makeSymbol(".");
		public static SymbolTerm Nil = SymbolTerm.makeSymbol("[]");
		Term arg1, arg2, arg3;

		public dollar_functor_3(Term a1, Term a2, Term a3, Predicate cont) {
			arg1 = a1; 
			arg2 = a2; 
			arg3 = a3;
			this.cont = cont;
		}
		public dollar_functor_3() {}
		public override void setArgument(Term[] args, Predicate cont){
			arg1 = args[0];
			arg2 = args[1];
			arg3 = args[2];
			this.cont = cont;
		}

		public override Predicate exec( Prolog engine ) {
			Term a1, a2, a3;
			a1 = arg1.dereference();
			a2 = arg2.dereference();
			a3 = arg3.dereference();

			Term functor, arity;
			Term[] args;

			if (a1.isVariable()) {
				if ( !a3.isInteger() ) {
					return engine.abort("{TYPE ERROR: " + this +
						" -arg 3: expected integer, found " +
						a3 + "}\n");
					//		return null;
				}
				int n = ((IntegerTerm)a3).value();
				if (n < 0) {
					return engine.abort("{DOMAIN ERROR: " + this +
						" -arg 3}\n");
					//		return null;
				}
				if (n == 0) {
					if ( !a2.isSymbol() &&  !a2.isInteger() ) {
						return engine.abort("{TYPE ERROR: " + this +
							" -arg 2: expected atomic, found " +
							a2 + "}\n");
						//		    return null;
					}
					if( !a1.unify(a2, engine.trail) )
						return engine.fail();
					return cont;
				}
				if ( !a2.isSymbol() ) {
					return engine.abort("{TYPE ERROR: " + this +
						" -arg 2: expected atom, found " +
						a2 + "}\n");
					//		return null;
				}
				if ( n == 2 && a2.Equals(SYM_dot) ) {
					//		Term t = new ListTerm(new VariableTerm(), new VariableTerm());
					Term t = new ListTerm(engine.makeVariable(), engine.makeVariable());
					if( !a1.unify(t, engine.trail) )
						return engine.fail();
					return cont;
				}
				args = new Term[n];
				for(int i=0; i<n; i++)
					args[i] = engine.makeVariable();
				//		args[i] = new VariableTerm();
				SymbolTerm sym = SymbolTerm.makeSymbol(((SymbolTerm)a2).name, n);
				if( !a1.unify(new StructureTerm(sym, args), engine.trail) )
					return engine.fail();
				return cont;
			} else {
				if ( a1.isSymbol() || a1.isInteger() ){
					functor = a1;
					arity   = new IntegerTerm(0);
				} else if ( a1.isList() ){
					functor = SYM_dot;
					arity   = new IntegerTerm(2);
				} else if ( a1.isStructure() ){
					//		functor = ((StructureTerm)a1).functor;
					functor = SymbolTerm.makeSymbol(((StructureTerm)a1).functor.name);
					arity   = new IntegerTerm(((StructureTerm)a1).arity);
				} else {
					return engine.fail();
				}
				if( !a2.unify(functor, engine.trail) ) 
					return engine.fail();
				if( !a3.unify(arity, engine.trail) ) 
					return engine.fail();
			}
			return cont;
		}

		public override int arity() { return 3; }

		public override string ToString() {
			return "$functor(" + arg1 + ", " + arg2 + ", " + arg3 + ")";
		}
	}

}