namespace JJC.Psharp.Lang {

	using System;
	using System.Threading;
	using System.Collections;

	public sealed class ConcurrentVariable {
		
		private ArrayList activeThreads =
			ArrayList.Synchronized( new ArrayList( ) );

		public CsObjectTerm csObject = null;

		public ConcurrentVariable( ) {
			csObject = new CsObjectTerm( this );
		}

		public void Pulse( ) {
			//lock( this.csObject ) 
			Monitor.PulseAll( this );
		}

		public bool AllWaiting( ) {
			foreach( ActiveThread at in activeThreads ) 
				if( !at.currentlyWaiting )
					return false;
			// now: all are waiting, so we mark them as pending release
			foreach( ActiveThread at in activeThreads )
				at.awaitingWaitRelease = true;
			Pulse( );
			return true;
		}

		public void DeclareWaiting( Prolog engine, bool waiting ) {
			ActiveThread found = null;
			foreach( ActiveThread at in activeThreads )
				if( at.threadDescriptor == engine )
					found = at;
			if( found != null ) {
				found.currentlyWaiting = waiting;
				lock( this )
					Pulse( );
			}
			//else
			//	throw new Exception( );
		}

		public bool ShouldRelease( Prolog engine ) {
			ActiveThread found = null;
			foreach( ActiveThread at in activeThreads )
				if( at.threadDescriptor == engine )
					found = at;
			if( found != null ) {
				if( found.awaitingWaitRelease ) {
					found.awaitingWaitRelease = false;
					lock( this )
						Pulse( );
					return true;
				} else
					return false;
			}
			//throw new Exception( );
			return false;
		}

		public void ReportThreadDeath( Prolog died ) {
			// died.current_output.WriteLine( "death: " + died );
			ActiveThread found = null;
			foreach( ActiveThread at in activeThreads )
				if( at.threadDescriptor == died )
					found = at;
			if( found != null ) {
				activeThreads.Remove( found );
				lock( this )
					Pulse( );
			}
		}

		public void ReportFork( Prolog born ) {
			// born.current_output.WriteLine( "birth: " + born );
			activeThreads.Add( new ActiveThread( born ) );
		}

		public override string ToString( ) {
			return "RootSynch: " + this.GetHashCode( );
		}
	}

	public sealed class ActiveThread {
		private bool awaitingWaitReleaseInternal = false;
		private bool currentlyWaitingInternal = false;
	
		public bool awaitingWaitRelease {
			set { 
				lock( this )
					awaitingWaitReleaseInternal = value;
			}
			get {
				lock( this )
					return awaitingWaitReleaseInternal;
			}
		}

		public bool currentlyWaiting {
			set { 
				lock( this )
					currentlyWaitingInternal = value;
			}
			get {
				lock( this )
					return currentlyWaitingInternal;
			}
		}

		public Prolog threadDescriptor = null;

		public ActiveThread( Prolog threadDescriptor ) {
			this.threadDescriptor = threadDescriptor;
		}			
	}

	public sealed class NewBinding {
		public CsObjectTerm vt;
		public Term b;
		public NewBinding( CsObjectTerm vt, Term b ) {
			this.vt = vt;
			this.b = b;
		}
	}

}