/*
 *  @(#)CpfEntry.cs
 *  Enrty of Choice Point Frame
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Collections;

	sealed class CpfEntry {
		public long timeStamp;
		public Term[] args;
		// public int length;
		public Predicate cont;  // continuation 
		public Predicate bp;    // next cluase
		public int tr;          // trail pointer
		public int b0;          // cut point
		public int R, L, U;
		public Term R1, R2;

		// for speed up...
		private CpfEntry next = null;
		//static private CpfEntry head = null;

		/* TEMP REPLACEMENT !!!!!!!!!!!!!!!! */

			/*
		public static CpfEntry makeCpfEntry(Term[] args, Predicate cont){
			return new CpfEntry(args, cont);
		}

		public void removeCpfEntry(){
		}  */

		/* TEMPORARAY CHANGE TO ASSESS MULTITHREADING CAPABILITY */

		public static CpfEntry makeCpfEntry(Term[] args, Predicate cont,
					CpfEntry head ){
			if (head == null){
				return new CpfEntry(args, cont);
			}
			CpfEntry entry = head;
			head = head.next;
			entry.args = args;
			entry.cont = cont;
			return entry;
		}

		public void removeCpfEntry( CpfEntry head ){
			this.args = null;
			this.next = head;
			head = this;
		}

		// */
		
		private CpfEntry(Term[] args, Predicate cont){
			this.args = args;
			this.cont = cont;
		}
	}

	/*
	   *  @(#)CpfStack.cs
	 *  Choice Point Stack
	 */

	/* 
	 * @author Mutsunori Banbara 
	 *         banbara@pascal.seg.kobe-u.ac.jp
	 *         Nara National College of Technology
	 * @author Naoyuki Tamura 
	 *         tamura@kobe-u.ac.jp
	 *         Kobe University
	 */


	public sealed class CpfStack {
		private int maxContents = 20000;
		private Stack buffer;
		private CpfEntry topEntry = null;
		private int itop;
		public int top {
			get {
				return itop;
			}
			set {
				itop = value;
			}
		}

		private CpfEntry head = null;

		public CpfStack( ){
			buffer = new Stack( maxContents ); // CpfEntry[maxContents];
			top = -1;
			head = null;
		}
		public CpfStack( int n ){
			maxContents = n;
			buffer = new Stack( maxContents ); // CpfEntry[maxContents];
			top = -1;
			head = null;
		}

		public void create(Term[] args, Predicate p){
			// try {
			top++;
			buffer.Push( topEntry = CpfEntry.makeCpfEntry(args, p, head) );
			// buffer[++top] = CpfEntry.makeCpfEntry(args, p);
			/*
			} catch (System.ArgumentOutOfRangeException) {
				Console.WriteLine( "{expanding choice point stack...}" );
				int len = buffer.Length;
				CpfEntry[] new_buffer = new CpfEntry[ len + 10000 ];
				// special array copy? JJC
				for( int i = 0; i < len; i++ ) {
					new_buffer[i] = buffer[i];
				}
				buffer = new_buffer;
				topEntry = CpfEntry.makeCpfEntry(args, p);
				// topEntry = new CpfEntry(args, p);
			}
			*/
		}

		public void delete( ) { 
			top--;
			(((CpfEntry)(buffer.Pop( )))).removeCpfEntry( head );
			topEntry = (CpfEntry)buffer.Peek( );
		}

		public void cut(int i){
			while (top > i) {
				top--;
				((CpfEntry)(buffer.Pop( ))).removeCpfEntry( head );
			}
			topEntry = (CpfEntry)buffer.Peek( );
		}

		public void setTimeStamp(long n){ topEntry.timeStamp = n; }
		public long getTimeStamp(){ return topEntry.timeStamp; }

		public void setBP(Predicate p){ topEntry.bp = p; }
		public void setTR(int i){ topEntry.tr = i; }
		public void setB0(int i){ topEntry.b0 = i; }
		public Term[] getArgs(){ return topEntry.args; }
		//    public int  getLength(){ return topEntry.length; }
		public Predicate getCont(){ return topEntry.cont; }
		public Predicate getBP(){ return topEntry.bp; }
		public int getTR(){ return topEntry.tr; }
		public int getB0(){ return topEntry.b0; }

		public Term getR1(){ return topEntry.R1; }
		public Term getR2(){ return topEntry.R2; }
		public void setR1(Term i){ topEntry.R1 = i; }
		public void setR2(Term i){ topEntry.R2 = i; }

		public int getR(){ return topEntry.R; }
		public void setR(int i){ topEntry.R = i; }
		public int getL(){ return topEntry.L; }
		public void setL(int i){ topEntry.L = i; }
		public int getU(){ return topEntry.U; }
		public void setU(int i){ topEntry.U = i; }

		public bool empty(){ return top == -1; }
		public int max(){ return maxContents; }
	}

}