
/*
 * @(#)Token.cs 
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

/*
   int read_token(StringBuilder mstring, PushbackReader stream)
     reads 1 token from the stream,
     sets the mstring, and returns the token type.

   Type		String
   -2		"error message"
   -1		"end_of_file"
   '.'		"."		full stop
   ' '		" "		space or comment or unknown chars
   ','		","
   '('		"("
   ')'		")"
   '['		"["
   ']'		"]"
   '{'		"{"
   '}'		"}"
   ','		","
   '|'		"|"
   'I'		"decimal"	positive integer
   'F'		"decimal"	positive float  //added by banbara at Dec. 1997
   'A'		"atom name"
   'V'		"variable name"
   'S'		"string"

   Floating point numbers and integers with radix are not supported.
*/

namespace JJC.Psharp.Lang {

	using System;
	using System.Text;
	using System.IO;

	public sealed class Token {

		public static string getReplacement( int c ) {
			switch (c) {
				case '+':  return "plus";
				case '-':  return "dash";
				case '*':  return "star";
				case '/':  return "slash";
				case '\\': return "bslash";
				case '^':  return "caret";
				case '<':  return "less";
				case '>':  return "gtr";
				case '=':  return "eq";
				case '`':  return "bquote";
				case '~':  return "tilde";
				case ':':  return "colon";
				case '.':  return "stop";
				case '?':  return "qn";
				case '@':  return "at";
				case '#':  return "hash";
				case '$':  return "dollar_";
				case '&':  return "amp";
				case '!':  return "bang";
				case ';':  return "scolon";
				case ',':  return "comma";
				default: throw new Exception( "Symbol conversion failed" );
			}
		}

		public static bool isSolo(int c) {
			return (c =='!' || c ==';');
		}

		public static bool isSymbol(int c) {
			switch (c) {
				case '+':
				case '-':
				case '*':
				case '/':
				case '\\':
				case '^':
				case '<':
				case '>':
				case '=':
				case '`':
				case '~':
				case ':':
				case '.':
				case '?':
				case '@':
				case '#':
				case '$':
				case '&':
					return true;
				default:
					return false;
			}
		}

		/*

		public static int read_token(StringBuilder mstring, PushbackReader stream) {

			int c, c1;
			int rc;

			c = stream.Peek();
		
			if(c == 26) {
				c = stream.Read();
				mstring.Append("end_of_file");
				return -1;
			}
		
			if (System.Char.IsDigit((char)c)) {
				rc = read_number(c, mstring, stream);
				if (rc == 1)
					rc = 'I';
				else if (rc == 2)
					rc = 'F';
				return rc;
			}

			if (System.Char.IsLower((char)c)) {
				rc = read_word(c, mstring, stream);
				if (rc > 0)
					rc = 'A';
				return rc;
			}

			if (System.Char.IsUpper((char) c) || c == '_') {
				rc = read_word(c, mstring, stream);
				if (rc > 0)
					rc = 'V';
				return rc;
			}

			switch (c) {
				case '(':
				case ')':
				case '[':
				case ']':
				case '{':
				case '}':
				case ',':
				case '|':
					stream.Read();
					mstring.Append((char) c);
					return c;
				case '.':		// full stop or symbol 
					stream.Read();
					mstring.Append(".");
					c1 = stream.Peek();
					if (c1 == -1 || c1 <= ' ') {
						stream.Read();
						return '.';
					}
					break;
				case '%':		// one line comment
					stream.Read();
					mstring.Append(" ");
					while ((c1 = stream.Read()) != '\n') {
						if(c1 == -1)
							return ' ';
					}
					return ' ';

				case '/':		// start of comment or symbol 
					stream.Read();
					if ((c1 = stream.Peek()) == '*') {
						stream.Read();
						mstring.Append(" ");
						while (true) {
							while ((c1 = stream.Read()) != '*') {
								if(c1 == -1) {
									mstring.Append("Unexpected end_of_file");
									return -2;
								}
							}
							if ((c1 = stream.Peek()) == '/') {
								stream.Read();
								return ' ';
							}
						}
					}
					break;

				case '\'':
					stream.Read();
					rc = read_quoted(c, mstring, stream);
					if (rc > 0)
						rc = 'A';
					return rc;
				case '"':
					stream.Read();
					rc = read_quoted(c, mstring, stream);
					if (rc > 0)
						rc = 'S';
					return rc;
				default:
					// stream.Read();
					break;
			}
			if (isSolo(c)) {
				stream.Read( );
				mstring.Append((char) c);
				return 'A';
			}
		
			if (c == '.') { 
				stream.Read( );
				return 'A';
			}
			if (isSymbol(c)) {
				rc = read_symbol(c, mstring, stream);
				if (rc > 0)
					rc = 'A';
				return rc;
			}
			stream.Read( );
			mstring.Append(" ");
			return ' ';
		}

		//---------------------------------------------------------------
		//  read_number(int c, StringBuilder s, PushbackReader stream)
		//---------------------------------------------------------------
		public static int read_number(int c, StringBuilder mstring, PushbackReader stream) {
			int c1;
			StringBuilder p = mstring;
			// stream.Unread(c);
			for (;;) {
				c1 = stream.Peek();
				if (!Char.IsDigit((char) c1))
					break;
				stream.Read();
				p.Append((char) c1);
			}
			return 1;
		}

		//---------------------------------------------------------------
		//  read_word(int c, StringBuilder s, PushbackReader stream)
		//---------------------------------------------------------------
		public static int read_word(int c, StringBuilder mstring, PushbackReader stream) {
			int c1;
			StringBuilder p = mstring;
	     
			// stream.Unread(c);
			for (;;) {
				c1 = stream.Peek();
				if (! Char.IsLetterOrDigit((char) c1) && c1 != '_')
					break;
				stream.Read();
				p.Append((char) c1);
			}
			return 1;
		}

		//---------------------------------------------------------------
		//  read_quoted(int q, StringBuilder s, PushbackReader stream)
		//---------------------------------------------------------------
		public static int read_quoted(int quote, StringBuilder mstring, PushbackReader stream)
		  {
			 int c1;
			 StringBuilder p = mstring;
	     
			 for (;;) {
			 c1 = stream.Peek();
			 if (c1 == -1 || c1 == '\n') {
				 p = new StringBuilder("Unexpected end_of_line");
				 return -2;
			 }
			stream.Read();
			if (c1 == quote) {
				 c1 = stream.Peek();
				 if (c1 != quote) {
				 return 1;
				 }
			 }
			 stream.Read();
			 p.Append((char) c1);
			 }
		}

		//---------------------------------------------------------------
		//  read_symbol(int c, StringBuilder s, PushbackReader stream) 
		//---------------------------------------------------------------
		public static int read_symbol(int c, StringBuilder mstring, PushbackReader stream) {

			int c1;
			StringBuilder p = mstring;
			p.Append((char) c);
			// stream.Unread(c); 
			for (;;) {
				c1 = stream.Peek();
				if (! isSymbol(c1))
					break;
				stream.Read();
				p.Append((char) c1);
			}
			return 1;
		}

		*/
	
		public static int 
			read_token(StringBuilder mstring, PushbackReader stream) {

			int c, c1;
			int rc;
			c = stream.Read();
			if(c == -1) {
				mstring.Append("end_of_file");
				return -1;
			}

			if (System.Char.IsDigit((char)c)) {
				rc = read_number(c, mstring, stream);
				if (rc == 1)
					rc = 'I';
				else if (rc == 2)
					rc = 'F';
				return rc;
			}

			if (System.Char.IsLower((char)c)) {
				rc = read_word(c, mstring, stream);
				if (rc > 0)
					rc = 'A';
				return rc;
			}

			if (System.Char.IsUpper((char) c) || c == '_') {
				rc = read_word(c, mstring, stream);
				if (rc > 0)
					rc = 'V';
				return rc;
			}

			switch (c) {
				case '(':
				case ')':
				case '[':
				case ']':
				case '{':
				case '}':
				case ',':
				case '|':
					mstring.Append((char) c);
					return c;
				case '.':		/* full stop or symbol */
					mstring.Append(".");
					c1 = stream.Read();
					if (c1 == -1 || c1 <= ' ') {
						c1 = stream.Read();
						if( !( c1 == -1 || c1 <= ' ' ) )
							stream.Unread( c1 );
						return '.';
					}
					stream.Unread(c1);
					break;
				case '%':		/* one line comment */
					mstring.Append(" ");
					while ((c1 = stream.Read()) != '\n') {
						if(c1 == -1)
							return ' ';
					}
					return ' ';

				case '/':		/* start of comment or symbol */
					if ((c1 = stream.Read()) == '*') {
						mstring.Append(" ");
						while (true) {
							while ((c1 = stream.Read()) != '*') {
								if(c1 == -1) {
									mstring.Append("Unexpected end_of_file");
									return -2;
								}
							}
							if ((c1 = stream.Read()) == '/')
								return ' ';
							stream.Unread(c1);
						}
					}
					stream.Unread(c1);
					break;

				case '\'':
					rc = read_quoted(c, mstring, stream);
					if (rc > 0)
						rc = 'A';
					return rc;
				case '"':
					rc = read_quoted(c, mstring, stream);
					if (rc > 0)
						rc = 'S';
					return rc;
				default:
					break;
			}
			if (isSolo(c)) {
				mstring.Append((char) c);
				return 'A';
			}
			// $B<h$j9g$($:!"$3$l$G2?$H$+N?$4$&!*(B
			if (c == '.') return 'A';
			if (isSymbol(c)) {
				rc = read_symbol(c, mstring, stream);
				if (rc > 0)
					rc = 'A';
				return rc;
			}
			mstring.Append(" ");
			return ' ';
		}

		/*---------------------------------------------------------------
		  read_number(int c, StringBuilder s, PushbackReader stream)
		---------------------------------------------------------------*/
		public static int read_number(int c, StringBuilder mstring, PushbackReader stream) {
			int c1;
			StringBuilder p = mstring;
			stream.Unread(c);
			for (;;) {
				c1 = stream.Read();
				if (!System.Char.IsDigit((char) c1))
					break;
				p.Append((char) c1);
			}
			stream.Unread(c1);
			return 1;
		}

		/*---------------------------------------------------------------
		  read_word(int c, StringBuilder s, PushbackReader stream)
		---------------------------------------------------------------*/
		public static int read_word(int c, StringBuilder mstring, PushbackReader stream) {
			int c1;
			StringBuilder p = mstring;
	     
			stream.Unread(c);
			for (;;) {
				c1 = stream.Read();
				if (! System.Char.IsLetterOrDigit((char) c1) && c1 != '_')
					break;
				p.Append((char) c1);
			}
			stream.Unread(c1);
			return 1;
		}

		/*---------------------------------------------------------------
		  read_quoted(int q, StringBuilder s, PushbackReader stream)
		---------------------------------------------------------------*/
		public static int read_quoted(int quote, StringBuilder mstring, PushbackReader stream) {
			int c1;
			StringBuilder p = mstring;
	     
			for (;;) {
				c1 = stream.Read();
				if (c1 == -1 || c1 == '\n') {
					stream.Unread(c1);
					p = new StringBuilder("Unexpected end_of_line");
					return -2;
				}
				if (c1 == quote) {
					c1 = stream.Read();
					if (c1 != quote) {
						stream.Unread(c1);
						return 1;
					}
				}
				p.Append((char) c1);
			}
		}

		/*---------------------------------------------------------------
		  read_symbol(int c, StringBuilder s, PushbackReader stream) 
		---------------------------------------------------------------*/
		public static int read_symbol(int c, StringBuilder mstring, PushbackReader stream) {

			int c1;
			StringBuilder p = mstring;
			p.Append((char) c);
			/* stream.Unread(c); */
			for (;;) {
				c1 = stream.Read();
				if (! isSymbol(c1))
					break;
				p.Append((char) c1);
			}
			stream.Unread(c1);
			return 1;
		}


		/*---------------------------------------------------------------
		  write_string(String s, PrintWriter p)
		---------------------------------------------------------------*/
		public static void 
			write_string(string mstring, TextWriter p) {
			p.Write(mstring);
			//	p.flush();
		}

		/*---------------------------------------------------------------
		  writeq_string(String s, PrintWriter p)
		---------------------------------------------------------------*/
		public static void 
			writeq_string(string mstring, TextWriter p) {
			char[] ch = mstring.ToCharArray();

			if ((string_type(mstring) == 3)) {
				p.Write("\'");
				for (int i=0; i<ch.Length; i++) {
					p.Write(ch[i]);
					if (ch[i] == '\'')
						p.Write("\'");
				}
				p.Write("\'");
				//	    p.flush();
			} else {
				write_string(mstring, p);
			}
		}

		/*---------------------------------------------------------------
		  string_type(String mstring)
		  0 : if mstring is a lower case alphnum
		  1 : if mstring is a symbol
		  2 : if mstring is a solo
		  3 : others
		---------------------------------------------------------------*/
		public static int 
			string_type(string mstring){

			char[] p = mstring.ToCharArray();
			if (mstring.Equals("[]") || mstring.Equals("{}"))
				return 0;
			if (mstring.Equals("") || mstring.Equals("."))
				return 3;
			if (mstring.Equals("!") || mstring.Equals(";"))
				return 2;
			if (Char.IsLower(p[0])) {
				for (int i=1; i<p.Length; i++) {
					if (! Char.IsLetterOrDigit(p[i]) && ((int)p[i]) != '_')
						return 3;
				}
				return 0;
			}

			if (isSymbol((int) p[0])) {
				for (int i=1; i<p.Length; i++) {
					if (! isSymbol((int) p[i]))
						return 3;
				}
				return 1;
			}
			return 3;
		}
	}

}