/*
 * *** Please do not edit ! ***
 * @(#) Write_1.cs
 * @procedure write/1 in inout.pl
 */

/*
 * @version P# 0.1, on March 18 2002;  Prolog Cafe 0.44, on November 12 1999
 * @author Mutsunori Banbara (banbara@pascal.seg.kobe-u.ac.jp)
 * @author Naoyuki Tamura    (tamura@kobe-u.ac.jp)
 * Modified by Jonathan Cook (jjc@dcs.ed.ac.uk)
 */
namespace JJC.Psharp.Predicates {

using JJC.Psharp.Lang;
using JJC.Psharp.Lang.Resource;
using Predicates = JJC.Psharp.Predicates;
using Resources = JJC.Psharp.Resources;

public class Write_1 : Predicate {
    static internal readonly Predicate Write_1_1 = new Predicates.Write_1_1();
    static internal readonly Predicate Write_1_2 = new Predicates.Write_1_2();
    static internal readonly Predicate Write_1_3 = new Predicates.Write_1_3();
    static internal readonly Predicate Write_1_4 = new Predicates.Write_1_4();
    static internal readonly Predicate Write_1_sub_1 = new Predicates.Write_1_sub_1();

    public Term arg1;

    public Write_1(Term a1, Predicate cont) {
        arg1 = a1; 
        this.cont = cont;
    }

    public Write_1(){}
    public override void setArgument(Term[] args, Predicate cont) {
        arg1 = args[0]; 
        this.cont = cont;
    }

    public override Predicate exec( Prolog engine ) {
        engine.aregs[1] = arg1;
        engine.cont = cont;
        return call( engine );
    }

    public virtual Predicate call( Prolog engine ) {
        engine.setB0();
        return engine.jtry(Write_1_1, Write_1_sub_1);
    }

    public override int arity() { return 1; }

    public override string ToString() {
        return "write(" + arg1 + ")";
    }
}

sealed class Write_1_sub_1 : Write_1 {
    static internal readonly Predicate Write_1_sub_2 = new Predicates.Write_1_sub_2();

    public override Predicate exec( Prolog engine ) {
        return engine.retry(Write_1_2, Write_1_sub_2);
    }
}

sealed class Write_1_sub_2 : Write_1 {
    static internal readonly Predicate Write_1_sub_3 = new Predicates.Write_1_sub_3();

    public override Predicate exec( Prolog engine ) {
        return engine.retry(Write_1_3, Write_1_sub_3);
    }
}

sealed class Write_1_sub_3 : Write_1 {

    public override Predicate exec( Prolog engine ) {
        return engine.trust(Write_1_4);
    }
}

sealed class Write_1_1 : Write_1 {

    public override Predicate exec( Prolog engine ) {
        Term a1, a2;
        Predicate p1, p2, p3;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        a2 = engine.makeVariable();
        p1 = new Predicates.dollar_fastWrite_1(a1, cont);
        p2 = new Predicates.dollar_cut_1(a2, p1);
        p3 = new Predicates.dollar_atomic_1(a1, p2);
        return new Predicates.dollar_getLevel_1(a2, p3);
    }
}

sealed class Write_1_2 : Write_1 {

    public override Predicate exec( Prolog engine ) {
        Term a1, a2;
        Predicate p1, p2, p3;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        a2 = engine.makeVariable();
        p1 = new Predicates.dollar_fastWrite_1(a1, cont);
        p2 = new Predicates.dollar_cut_1(a2, p1);
        p3 = new Predicates.dollar_csObject_1(a1, p2);
        return new Predicates.dollar_getLevel_1(a2, p3);
    }
}

sealed class Write_1_3 : Write_1 {
    static internal readonly SymbolTerm s1 = SymbolTerm.makeSymbol("numbervars");
    static internal readonly SymbolTerm s2 = SymbolTerm.makeSymbol("[]");
    static internal readonly ListTerm s3 = new ListTerm(s1, s2);

    public override Predicate exec( Prolog engine ) {
        Term a1, a2;
        Predicate p1, p2;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        a2 = engine.makeVariable();
        p1 = new Predicates.dollar_cut_1(a2, cont);
        p2 = new Predicates.dollar_writeTerm_2(a1, s3, p1);
        return new Predicates.dollar_getLevel_1(a2, p2);
    }
}

sealed class Write_1_4 : Write_1 {

    public override Predicate exec( Prolog engine ) {
        return engine.cont;
    }
}
}

