/*
 * *** Please do not edit ! ***
 * @(#) dollar_leash_1.cs
 * @procedure $leash/1 in interpreter.pl
 */

/*
 * @version P# 0.1, on March 18 2002;  Prolog Cafe 0.44, on November 12 1999
 * @author Mutsunori Banbara (banbara@pascal.seg.kobe-u.ac.jp)
 * @author Naoyuki Tamura    (tamura@kobe-u.ac.jp)
 * Modified by Jonathan Cook (jjc@dcs.ed.ac.uk)
 */
namespace JJC.Psharp.Predicates {

using JJC.Psharp.Lang;
using JJC.Psharp.Lang.Resource;
using Predicates = JJC.Psharp.Predicates;
using Resources = JJC.Psharp.Resources;

public class dollar_leash_1 : Predicate {
    static internal readonly Predicate dollar_leash_1_1 = new Predicates.dollar_leash_1_1();
    static internal readonly Predicate dollar_leash_1_2 = new Predicates.dollar_leash_1_2();
    static internal readonly Predicate dollar_leash_1_3 = new Predicates.dollar_leash_1_3();
    static internal readonly Predicate dollar_leash_1_4 = new Predicates.dollar_leash_1_4();
    static internal readonly Predicate dollar_leash_1_var = new Predicates.dollar_leash_1_var();

    public Term arg1;

    public dollar_leash_1(Term a1, Predicate cont) {
        arg1 = a1; 
        this.cont = cont;
    }

    public dollar_leash_1(){}
    public override void setArgument(Term[] args, Predicate cont) {
        arg1 = args[0]; 
        this.cont = cont;
    }

    public override Predicate exec( Prolog engine ) {
        engine.aregs[1] = arg1;
        engine.cont = cont;
        return call( engine );
    }

    public virtual Predicate call( Prolog engine ) {
        engine.setB0();
        return engine.switch_on_term(
                                   dollar_leash_1_var,
                                   dollar_leash_1_4,
                                   dollar_leash_1_var,
                                   dollar_leash_1_4,
                                   dollar_leash_1_4
                                   );
    }

    public override int arity() { return 1; }

    public override string ToString() {
        return "$leash(" + arg1 + ")";
    }
}

sealed class dollar_leash_1_var : dollar_leash_1 {
    static internal readonly Predicate dollar_leash_1_var_1 = new Predicates.dollar_leash_1_var_1();

    public override Predicate exec( Prolog engine ) {
        return engine.jtry(dollar_leash_1_1, dollar_leash_1_var_1);
    }
}

sealed class dollar_leash_1_var_1 : dollar_leash_1 {
    static internal readonly Predicate dollar_leash_1_var_2 = new Predicates.dollar_leash_1_var_2();

    public override Predicate exec( Prolog engine ) {
        return engine.retry(dollar_leash_1_2, dollar_leash_1_var_2);
    }
}

sealed class dollar_leash_1_var_2 : dollar_leash_1 {
    static internal readonly Predicate dollar_leash_1_var_3 = new Predicates.dollar_leash_1_var_3();

    public override Predicate exec( Prolog engine ) {
        return engine.retry(dollar_leash_1_3, dollar_leash_1_var_3);
    }
}

sealed class dollar_leash_1_var_3 : dollar_leash_1 {

    public override Predicate exec( Prolog engine ) {
        return engine.trust(dollar_leash_1_4);
    }
}

sealed class dollar_leash_1_1 : dollar_leash_1 {
    static internal readonly SymbolTerm s1 = SymbolTerm.makeSymbol("[]");
    static internal readonly SymbolTerm f2 = SymbolTerm.makeSymbol("$current_leash", 1);
    static internal readonly SymbolTerm s4 = SymbolTerm.makeSymbol("{No leashing}");

    public override Predicate exec( Prolog engine ) {
        Term a1, a2;
        Predicate p1, p2, p3;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        if ( !s1.unify(a1, engine.trail) ) return engine.fail();
        Term[] h3 = {engine.makeVariable()};
        a2 = new StructureTerm(f2, h3);
        p1 = new Predicates.dollar_nl_0(cont);
        p2 = new Predicates.Write_1(s4, p1);
        p3 = new Predicates.Retractall_1(a2, p2);
        return new Predicates.dollar_neckCut_0(p3);
    }
}

sealed class dollar_leash_1_2 : dollar_leash_1 {
    static internal readonly SymbolTerm s1 = SymbolTerm.makeSymbol("none");
    static internal readonly SymbolTerm s2 = SymbolTerm.makeSymbol("[]");

    public override Predicate exec( Prolog engine ) {
        Term a1;
        Predicate p1;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        if ( !s1.unify(a1, engine.trail) ) return engine.fail();
        p1 = new Predicates.dollar_leash_1(s2, cont);
        return new Predicates.dollar_neckCut_0(p1);
    }
}

sealed class dollar_leash_1_3 : dollar_leash_1 {
    static internal readonly SymbolTerm s1 = SymbolTerm.makeSymbol("full");
    static internal readonly SymbolTerm f2 = SymbolTerm.makeSymbol("$current_leash", 1);
    static internal readonly SymbolTerm s4 = SymbolTerm.makeSymbol("{Leashing stopping on ");
    static internal readonly SymbolTerm s5 = SymbolTerm.makeSymbol("}");

    public override Predicate exec( Prolog engine ) {
        Term a1, a2, a3;
        Predicate p1, p2, p3, p4, p5, p6, p7;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        if ( !s1.unify(a1, engine.trail) ) return engine.fail();
        Term[] h3 = {engine.makeVariable()};
        a2 = new StructureTerm(f2, h3);
        a3 = engine.makeVariable();
        p1 = new Predicates.dollar_nl_0(cont);
        p2 = new Predicates.Write_1(s5, p1);
        p3 = new Predicates.Write_1(a3, p2);
        p4 = new Predicates.Write_1(s4, p3);
        p5 = new Predicates.dollar_assertLeashes_1(a3, p4);
        p6 = new Predicates.dollar_fullLeash_1(a3, p5);
        p7 = new Predicates.Retractall_1(a2, p6);
        return new Predicates.dollar_neckCut_0(p7);
    }
}

sealed class dollar_leash_1_4 : dollar_leash_1 {
    static internal readonly SymbolTerm f1 = SymbolTerm.makeSymbol("$current_leash", 1);
    static internal readonly SymbolTerm s3 = SymbolTerm.makeSymbol("{Leashing stopping on ");
    static internal readonly SymbolTerm s4 = SymbolTerm.makeSymbol("}");

    public override Predicate exec( Prolog engine ) {
        Term a1, a2;
        Predicate p1, p2, p3, p4, p5;
        a1 = engine.aregs[1].dereference();
        Predicate cont = engine.cont;

        Term[] h2 = {engine.makeVariable()};
        a2 = new StructureTerm(f1, h2);
        p1 = new Predicates.dollar_nl_0(cont);
        p2 = new Predicates.Write_1(s4, p1);
        p3 = new Predicates.Write_1(a1, p2);
        p4 = new Predicates.Write_1(s3, p3);
        p5 = new Predicates.dollar_assertLeashes_1(a1, p4);
        return new Predicates.Retractall_1(a2, p5);
    }
}
}

