/*
 *  @(#)ClosureTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

// import java.lang.*;
// import java.util.*;

namespace JJC.Psharp.Lang.Resource {

	using JJC.Psharp.Lang;

	public sealed class ClosureTerm : Term {

		private Predicate icode;

		public Predicate code {
			get {
				return icode;
			}
			set {
				icode = value;
			}
		}

		private Term[] ifreeVars;
		public Term[] freeVars {
			get {
				return ifreeVars;
			}
			set {
				ifreeVars = value;
			}
		}
		private int ilength;
		public int length {
			get {
				return ilength;
			}
			set {
				ilength = value;
			}
		}

		public ClosureTerm(Predicate _code, Term[] _freeVars) {
			code = _code;
			freeVars = _freeVars;
			length = freeVars.Length;
		}

		public override bool unify(Term t, Trail trail) {
			t = t.dereference();
			if ( t.isVariable() )
				return ((VariableTerm)t).unify(this, trail);
			if ( !t.isClosure() )
				return false;
			ClosureTerm s = (ClosureTerm)t;
			if ( ! code.Equals(s.code) )
				return false;
			if (  length != s.length )
				return false;
			for (int i=0; i<length; i++)
				if ( !freeVars[i].unify(s.freeVars[i], trail) )
					return false;
			return true;
		}

		public override bool Equals(Term t) {
			Term a1,a2;
			//	t = t.dereference();
			if ( !t.isClosure() )
				return false;
			ClosureTerm s = (ClosureTerm)t;
			if ( ! code.Equals(s.code) )
				return false;
			if (  length != s.length )
				return false;
			for ( int i=0; i<length; i++ ){
				a1 = freeVars[i].dereference();
				a2 = s.freeVars[i].dereference();
				if ( ! a1.Equals(a2) )
					return false;
			}
			return true;
		}

		/*
		public Term copyTerm(Hashtable table) {
		Term[] a = new Term[length];
		for (int i = 0; i < length; i++)
			a[i] = freeVars[i].copyTerm(table);
		return new ClosureTerm(code, a);
		}
		*/

		public override Term copyTerm(Prolog engine) {
			Term[] a = new Term[length];
			for (int i = 0; i < length; i++)
				a[i] = freeVars[i].copyTerm(engine);
			return new ClosureTerm(code, a);
		}

		public override string ToString() {
			string s;
			string delim = "";
			s = "closure(" + code.ToString() + ", [";
			for (int i = 0; i < length; i++) {
				s += delim + freeVars[i].ToString();
				delim = ",";
			}
			s += "])";
			return s;
		}
	}

}

